// Copyright (C) 2007 Juan Manuel Borges Caño

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#include <wl.h>
#include <X11/Xlib.h>
#include <GL/glx.h>
#include <stdio.h>
#include <stdlib.h>

static Display *_dpy;
static Window _win;
static GLXContext _ctx;
static Cursor _blank_cursor;
static void (*_uDraw)(void *) = NULL;
static void *_uDrawData = NULL;
static void (*_uReshape)(WLsizei, WLsizei, void *) = NULL;
static void *_uReshapeData = NULL;

/*
 * Create an RGB, double-buffered window.
 * Return the window and context handles.
 */
static void
make_window(Display *_dpy, const char *name, int x, int y, int width, int height, Window *winRet, GLXContext *ctxRet)
{
	int attrib[] =
	{
		GLX_RGBA,
		GLX_RED_SIZE, 1,
		GLX_GREEN_SIZE, 1,
		GLX_BLUE_SIZE, 1,
		GLX_DOUBLEBUFFER,
		GLX_DEPTH_SIZE, 1,
		None
       	};
	int scrnum;
	XSetWindowAttributes attr;
	unsigned long mask;
	Window root;
	Window win;
	GLXContext ctx;
	XVisualInfo *visinfo;
	scrnum = DefaultScreen(_dpy);
	root = RootWindow(_dpy, scrnum);
	visinfo = glXChooseVisual(_dpy, scrnum, attrib);
	if (!visinfo)
	{
		fprintf(stderr, "Couldn't get an RGB, Double-buffered visual\n");
		exit(EXIT_FAILURE);
	}
	/* window attributes */
	attr.background_pixel = 0;
	attr.border_pixel = 0;
	attr.colormap = XCreateColormap(_dpy, root, visinfo->visual, AllocNone);
	attr.event_mask = StructureNotifyMask | ExposureMask | KeyPressMask;
	mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;
	win = XCreateWindow(_dpy, root, 0, 0, width, height, 0, visinfo->depth, InputOutput, visinfo->visual, mask, &attr);
	/* set hints and properties */
	{
		XSizeHints sizehints;
		sizehints.x = x;
		sizehints.y = y;
		sizehints.width  = width;
		sizehints.height = height;
		sizehints.flags = USSize | USPosition;
		XSetNormalHints(_dpy, win, &sizehints);
		XSetStandardProperties(_dpy, win, name, name, None, (char **)NULL, 0, &sizehints);
	}
	ctx = glXCreateContext( _dpy, visinfo, NULL, True );
	if (!ctx)
	{
	      fprintf(stderr, "glXCreateContext failed\n");
	      exit(EXIT_FAILURE);
	}
	XFree(visinfo);
	*winRet = win;
	*ctxRet = ctx;
}

void
wlCreate(const char *title, WLsizei width, WLsizei height)
{
	if(!_uDraw)
	{
		fprintf(stderr, "Draw function not defined\n");
		exit(EXIT_FAILURE);
	}
	if(!_uReshape)
	{
		fprintf(stderr, "Reshape function not defined\n");
		exit(EXIT_FAILURE);
	}
	char *dpyName = NULL;
	_dpy = XOpenDisplay(dpyName);
	if (!_dpy)
	{
		fprintf(stderr, "couldn't open display %s\n", dpyName);
		exit(EXIT_FAILURE);
	}
	make_window(_dpy, title, 0, 0, width, height, &_win, &_ctx);
	XMapWindow(_dpy, _win);
	glXMakeCurrent(_dpy, _win, _ctx);

	/* make a blank cursor */
	Pixmap blank;
	XColor dummy;
	char data[1] = {0};
	blank = XCreateBitmapFromData (_dpy, _win, data, 1, 1);
	if(blank == None) fprintf(stderr, "error: out of memory.\n");
	_blank_cursor = XCreatePixmapCursor(_dpy, blank, blank, &dummy, &dummy, 0, 0);
	XFreePixmap (_dpy, blank);

	_uReshape(width, height, _uReshapeData);
}

void
wlDestroy()
{
	glXDestroyContext(_dpy, _ctx);
	XDestroyWindow(_dpy, _win);
	XCloseDisplay(_dpy);
}

void
wlSwapBuffers()
{
      glXSwapBuffers(_dpy, _win);
}

void
wlDrawFunc(void (*uDraw)(void *), void *data)
{
	_uDraw = uDraw;
	_uDrawData = data;
}

void
wlReshapeFunc(void (*uReshape)(WLsizei, WLsizei, void *), void *data)
{
	_uReshape = uReshape;
	_uReshapeData = data;
}

void
wlUpdate()
{
	while (XPending(_dpy) > 0)
	{
		XEvent event;
		XNextEvent(_dpy, &event);
		switch (event.type)
		{
			case ConfigureNotify:
				_uReshape(event.xconfigure.width, event.xconfigure.height, _uReshapeData);
				break;
		}
	}
	_uDraw(_uDrawData);
}

void
wlGrabPointer()
{
	XGrabPointer(_dpy, _win, True, 0, GrabModeAsync, GrabModeAsync, _win, _blank_cursor, CurrentTime);
}


void
wlUngrabPointer()
{
	XUngrabPointer(_dpy, CurrentTime);
}

WLboolean
wlHasFocus()
{
	Window focus_return;
	int revert_to_return;
	XGetInputFocus(_dpy, &focus_return, &revert_to_return);
	return (focus_return == _win);
}

void
wlGetBooleanv(WLenum pname, WLboolean *params)
{
	switch(pname)
	{
		case WL_FOCUS_BIT:
		{
			Window focus_return;
			int revert_to_return;
			XGetInputFocus(_dpy, &focus_return, &revert_to_return);
			*params = (focus_return == _win);
			break;
		}
	}
}

void
wlGetDoublev(WLenum pname, WLdouble *params)
{
}

void
wlGetFloatv(WLenum pname, WLfloat *params)
{
}

void
wlGetIntegerv(WLenum pname, WLint *params)
{
}
